package dev.pfaff.altargui.pathfinder;

import java.lang.foreign.Arena;
import java.lang.foreign.MemorySegment;

import static dev.pfaff.altargui.pathfinder.Pathfinder.*;
import static java.lang.foreign.MemoryLayout.PathElement.groupElement;
import static java.lang.foreign.ValueLayout.*;

public record PFGLRendererRef(MemorySegment ptr) {
	public static PFGLRendererRef create(PFGLDeviceRef device, PFResourceLoaderRef resourceLoader,
																		PFRendererMode mode, PFGLDestFramebufferRef framebuffer) {
		try (var arena = Arena.ofConfined()) {
			var options = arena.allocate(ML_PFRendererOptions);
			options.set(ML_PFDestFramebufferRef, ML_PFRendererOptions.byteOffset(groupElement(0)), framebuffer.ptr());
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("r")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("g")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("b")), 1f);
			options.set(JAVA_FLOAT, ML_PFRendererOptions.byteOffset(groupElement(1), groupElement("a")), 1f);
			// 0b1 : enable background color
			// 0b10 : enable debug ui
			options.set(JAVA_BYTE, ML_PFRendererOptions.byteOffset(groupElement(2)), (byte) 0b10);
			var ptr = (MemorySegment) MH_PFGLRendererCreate.invokeExact(device.ptr(),
					resourceLoader.ptr(),
					arena.allocateFrom(ML_PFRendererMode, mode.value),
//					arena.allocateFrom(ADDRESS, options)));
					options);
			return new PFGLRendererRef(ptr.reinterpret(Arena.ofAuto(), ptr_ -> {
				try {
					MH_PFGLRendererDestroy.invokeExact(ptr_);
				} catch (Throwable e) {
					throw new RuntimeException(e);
				}
			}));
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void destFramebufferSizeChanged() {
		try {
			MH_PFGLRendererDestFramebufferSizeChanged.invokeExact(this.ptr);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setViewport(float x, float y, float width, float height) {
		try (var arena = Arena.ofConfined()) {
//			var rect = arena.allocateFrom(JAVA_FLOAT, x, y, x + width, y + height).asSlice(0, ML_PFRectF);
			var rect = arena.allocateFrom(JAVA_FLOAT, x, y, width, height).asSlice(0, ML_PFRectF);
			if ((int) MH_PFGLRendererSetViewport.invokeExact(this.ptr, rect) != 0) {
				throw new RuntimeException();
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void setWindowSize(int width, int height) {
		try (var arena = Arena.ofConfined()) {
			var size = arena.allocateFrom(JAVA_INT, width, height).asSlice(0, ML_PFVector2I);
			if ((int) MH_PFGLRendererSetWindowSize.invokeExact(this.ptr, size) != 0) {
				throw new RuntimeException();
			}
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
