package dev.pfaff.altargui.render.objects;

import dev.pfaff.altargui.CanvasImpl;
import dev.pfaff.altargui.GuiGlobals;
import dev.pfaff.altargui.internal.Config;
import dev.pfaff.altargui.render.*;
import dev.pfaff.glass.opengl.enums.GLBufferAccess;
import dev.pfaff.glass.opengl.enums.GLBufferTarget;

import java.lang.foreign.MemoryLayout;
import java.util.function.Consumer;

/**
 * Might be optimal for large amounts of data. Testing will be needed to confirm this and establish a rough threshold.
 */
public record StaticGPUBuffer(int program, int id, MemoryLayout instanceLayout, int instanceCount) implements RenderObject {
	public static StaticGPUBuffer build(int layer, MemoryLayout instanceLayout, int instanceCount, Consumer<? super Sink> builder) {
		int length = (int)instanceLayout.byteSize() * instanceCount;

		int id = GLUtil.glGenBuffers();
		GuiGlobals.GL.glBindBuffer(GLBufferTarget.ARRAY_BUFFER, id);

		var buffer = GLUtil.glMapBuffer(GLBufferTarget.ARRAY_BUFFER, GLBufferAccess.WRITE_ONLY, length);
		var sink = new MemorySegmentSink(buffer);

		builder.accept(sink);

		if (sink.offset() != length) {
			throw new RuntimeException(STR."Buffer size not equal to allocated size: expected \{length}, found \{sink.offset()}");
		}

		GLUtil.glUnmapBuffer(GLBufferTarget.ARRAY_BUFFER);

		if (Config.GL_ALWAYS_UNBIND) GuiGlobals.GL.glBindBuffer(GLBufferTarget.ARRAY_BUFFER, 0);

		return new StaticGPUBuffer(layer, id, instanceLayout, instanceCount);
	}

	@Override
	public void close() {
		GLUtil.glDeleteBuffers(id);
	}

	@Override
	public void uploadTo(Sink sink) {
		sink.accept(this);
	}

	@Override
	public void uSpecUploadAtUnchecked(ViaIntermediateBuffer buffer, long offset) {
		RenderObject.super.uSpecUploadAtUnchecked(buffer, offset);
	}

	@Override
	public void uSpecUploadCanvas(CanvasImpl canvas) {
		RenderObject.super.uSpecUploadCanvas(canvas);
	}

	@Override
	public void uSpecUploadLayered(LayeredRenderBuffer buffer) {
		RenderObject.super.uSpecUploadLayered(buffer);
	}
}
